/*-----------------------------------------------------------------
  All code is provided for EDUCATIONAL purposes ONLY. There is no
  warranty and no support.

  Copyright (c) Jack Botner and licensed under the MIT license; see
  http://opensource.org/licenses/mit-license.php

  appltimer by Jack Botner

  appltimer.c

  This program implements an appliance timer with the following hardware:
  - a KTM-S1201 12 x 1 serial LCD display
  - a CTS rotary encoder with pushbutton
  - a relay that controls the appliance
  - NJU6355 real time clock

  Processor: ATMega88 (or ATMega168)
      Clock: Internal, 8.0 MHz

  20121101 Add support for changing the time when DST kicks
           in or out.
-----------------------------------------------------------------*/
#include <stdint.h>
#include <avr/io.h>
//#include <avr/wdt.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <util/delay_basic.h>
#include <stdlib.h>
#include <string.h>
#include "appltimer.h"
#include "calendar.h"
#include "ktms1201.h"
#include "nju6355.h"

#define NR_MAIN		7				// number of main menu items

// Global Variables

RTCPARMS       stRtc;
MY_EEPROM_DATA stEepromVars;		// variables kept in EEPROM

volatile uint8_t ucEventFlags1=0;		// program event flags 1
volatile uint8_t ucEventFlags2=0;		// program event flags 2
volatile uint8_t ucDebouncedState=0;	// debounced state of the switches
volatile uint8_t wasDST;				// previous DST state
volatile uint16_t uiMsPressed=0;		// how long encoder button pressed
volatile uint16_t uiSecondCtr=0;		// second counter, for LCD maintenance

static uint8_t ucMode=MODE_RUN;			// program mode
static uint8_t ucTempOverrideOn=0;		// user wants the appliacnce on

static uint8_t	ucPrepgmHoursOn, ucPrepgmHoursOff;
static uint8_t	ucPrepgmMinsOn,  ucPrepgmMinsOff;

MY_EEPROM_DATA stData;

// Internal Functions

static void check_DST( void );
static uint8_t check_on_or_off( void );
static void format_and_display_time(void);
static uint8_t main_menu(void);
static void set_current_programmed_on_off_times(void);
static void system_init(void);

/*-----------------------------------------------------------------
  main()
-----------------------------------------------------------------*/
int main( void )
{
  uint8_t ii, ucIndex, ucOn;

  system_init();          		// initialize system

  set_bit( PORTD, PORTD6 );		// power up the LCD
  for ( ii=0; ii<2; ++ii )		// 50 ms
	_delay_loop_2( 50000 );		// 25 ms
  LCD_init();					// initialize the LCD
  LCD_puts2( "Timer 1.0", 0 );
  
  rtc_init();					// initialize the rtc

  // Get the saved state variables from eeprom
  eeprom_read_block( &stData, 0, sizeof( MY_EEPROM_DATA ) );
  if ( stData.ucOnTimeHH > 23 ||		// probably EEPROM not initialized
	   stData.ucOnTimeMM > 59 ||
	   stData.ucOffTimeHH > 23 ||
	   stData.ucOffTimeMM > 59 )
  {
    stData.ucOnTimeHH = stData.ucOnTimeMM = 0;
    stData.ucOffTimeHH = stData.ucOffTimeMM = 0;
    stData.ucActivated = ACTIVATED_NO;
	stData.ucPreprogram = ACTIVATED_NO;
	set_bit( ucEventFlags1, EVENT_UPDATE_EEPROM );
  }

  // Initialize the preprogrammed on/off times
  set_current_programmed_on_off_times();

  // Initialize the DST flag
  wasDST = isdaylighttime( stRtc.rtc_year, stRtc.rtc_month, 
                           stRtc.rtc_day,  stRtc.rtc_hours );

  for ( ii=0; ii<80; ++ii )		// 2 s
	_delay_loop_2( 50000 );		// 25 ms

  ucEventFlags1 = 0;				// clear all events

  while(1)
  {
    if ( ucMode == MODE_RUN )
	{
	  // In RUN mode we display the time and watch for encoder events
	  // and see if we have to turn the appliance on or off.

 	  if ( test_bit( ucEventFlags1, EVENT_BUTTON_PRESSED ) )
	  {
	    ucMode = MODE_PROGRAM;
	    clear_bit( ucEventFlags1, EVENT_BUTTON_PRESSED );
	  }

	  if ( test_bit( ucEventFlags1, EVENT_ENCODER_UP ) ||
		   test_bit( ucEventFlags1, EVENT_ENCODER_DOWN ) )
	  {
	    // In run mode the encoder rotation is used to 
		// turn the device on or off.
		ucTempOverrideOn = ( ucTempOverrideOn ) ? 0 : 1;

		clear_bit( ucEventFlags1, EVENT_ENCODER_UP );
		clear_bit( ucEventFlags1, EVENT_ENCODER_DOWN );
	  }

	  if ( test_bit( ucEventFlags1, EVENT_READ_RTC ) )
	  {
		rtc_read();
		format_and_display_time();
		clear_bit( ucEventFlags1, EVENT_READ_RTC );
	  }

	  if ( test_bit( ucEventFlags1, EVENT_UPDATE_EEPROM ) )
	  {
		eeprom_write_block( &stData, 0, sizeof( MY_EEPROM_DATA ) );
		clear_bit( ucEventFlags1, EVENT_UPDATE_EEPROM );
	  }

	  if ( test_bit( ucEventFlags1, EVENT_ANALYZE_STATE ) )
	  {
		ucOn = check_on_or_off();
		if ( ucOn )
		{
		  // If the relay is off, turn it on
		  if ( !test_bit( RELAY_PORT_IN, RELAY_PIN ) )
		    set_bit( RELAY_PORT_OUT, RELAY_PIN );
		}
		else
		{
		  // If the relay is on, turn it off
		  if ( test_bit( RELAY_PORT_IN, RELAY_PIN ) )
		    clear_bit( RELAY_PORT_OUT, RELAY_PIN );
		}

		check_DST();

		clear_bit( ucEventFlags1, EVENT_ANALYZE_STATE );
	  }

	  if ( test_bit( ucEventFlags2, EVENT_LCD_MAINT ) )
	  {
	    // I do this because the LCD seems to conk out randomly after
		// a few hours/days/weeks.
		clear_bit( PORTD, PORTD6 );		// power down the LCD
		_delay_loop_2( 50000 );			// 25 ms
		set_bit( PORTD, PORTD6 );		// power up the LCD
		_delay_loop_2( 50000 );			// 25 ms
		LCD_init();						// initialize the LCD

		uiSecondCtr = 0;
		clear_bit( ucEventFlags2, EVENT_LCD_MAINT );
	  }
	}
	else	// MODE_PROGRAM
	{
	  // In PROGRAM mode we display the menu and perform user requests.

	  ucIndex = main_menu();	// returns index of menu item to process
	  switch ( ucIndex )
	  {
	    case INDEX_RUN:
		  // Go back to RUN mode
	      ucMode = MODE_RUN;
		  break;

	    case INDEX_ACTIVATE:
		  // Activate or deactivate the timer function and return to RUN mode.
		  if ( stData.ucActivated == ACTIVATED_NO )
		    stData.ucActivated = ACTIVATED_YES;
		  else
		    stData.ucActivated = ACTIVATED_NO;
		  set_bit( ucEventFlags1, EVENT_UPDATE_EEPROM );
	      ucMode = MODE_RUN;
		  break;

	    case INDEX_SETCLK:
		  // Set the RTC date and time.
		  program_clock();
	      ucMode = MODE_RUN;
		  break;

	    case INDEX_SETON:
		  // Set the timer ON time
		  program_time( ONTIME );
		  break;

	    case INDEX_SETOFF:
		  // Set the timer OFF time
		  program_time( OFFTIME );
		  break;

	    case INDEX_PREPGM:
		  // Enable or disable the special preprogram.
		  if ( stData.ucPreprogram == ACTIVATED_NO )
		    stData.ucPreprogram = ACTIVATED_YES;
		  else
		    stData.ucPreprogram = ACTIVATED_NO;
		  set_bit( ucEventFlags1, EVENT_UPDATE_EEPROM );
		  break;

	    case INDEX_CLKTEST:
		  // The clock test outputs a signal which can be
		  // measured to check the accuracy of the clock.
		  // You may want to disable it except for testing.
	      rtc_test();		// this function does not return.
	  }
	}
  }
}

/*-----------------------------------------------------------------
  main_menu()

  Displays the main menu.

  Returns: index of selected menu item.
-----------------------------------------------------------------*/
static uint8_t main_menu()
{
  static volatile uint8_t ucIndex=0;

  static char *msgs[NR_MAIN] = 
  {
    "Run", NULL, "Set Clock", "Set ON time", "Set OFF time", 
	NULL,  "CLK TEST"
  };

  while ( 1 )
  {
	if ( ucIndex == INDEX_ACTIVATE )
	  msgs[INDEX_ACTIVATE] = ( stData.ucActivated == ACTIVATED_YES ) ? "Timer Off" : "Timer On";

	if ( ucIndex == INDEX_PREPGM )
	  msgs[INDEX_PREPGM] = ( stData.ucPreprogram == ACTIVATED_YES ) ? "Prepgm Off" : "Prepgm On";
	
	LCD_puts2( msgs[ucIndex], 0 );

 	if ( test_bit( ucEventFlags1, EVENT_BUTTON_PRESSED ) )
	{
	  clear_bit( ucEventFlags1, EVENT_BUTTON_PRESSED );
	  return ucIndex;
	}

 	if ( test_bit( ucEventFlags1, EVENT_ENCODER_UP ) )
	{
	  ++ucIndex;
	  if ( ucIndex == NR_MAIN )
	    ucIndex = 0;
	  clear_bit( ucEventFlags1, EVENT_ENCODER_UP );
	  continue;
	}

 	if ( test_bit( ucEventFlags1, EVENT_ENCODER_DOWN ) )
	{
	  if ( ucIndex )
		--ucIndex;
	  else
	    ucIndex = NR_MAIN - 1;
	  clear_bit( ucEventFlags1, EVENT_ENCODER_DOWN );
	}
  }
}

/*-----------------------------------------------------------------
  format_and_display_time()

  Display the time on the LCD: HH MM SS
-----------------------------------------------------------------*/
static void format_and_display_time()
{
  char	buffer[12], temp[4];

  utoa( (uint16_t) stRtc.rtc_hours, temp, 10 );
  if ( strlen( temp ) == 1 )
  {
	buffer[0] = '0';
	buffer[1] = temp[0];
  }
  else
  {
	buffer[0] = temp[0];
	buffer[1] = temp[1];
  }
  buffer[2] = ' ';

  utoa( (uint16_t) stRtc.rtc_mins, temp, 10 );
  if ( strlen( temp ) == 1 )
  {
	buffer[3] = '0';
	buffer[4] = temp[0];
  }
  else
  {
	buffer[3] = temp[0];
	buffer[4] = temp[1];
  }
  buffer[5] = ' ';

  utoa( (uint16_t) stRtc.rtc_secs, temp, 10 );
  if ( strlen( temp ) == 1 )
  {
	buffer[6] = '0';
	buffer[7] = temp[0];
  }
  else
  {
	buffer[6] = temp[0];
	buffer[7] = temp[1];
  }

  buffer[8] = '\0';
  LCD_puts( buffer, 1 );

  if ( stData.ucActivated == ACTIVATED_YES )
  {
	LCD_insert_decimal( 10 );

	if ( stData.ucPreprogram == ACTIVATED_YES )
	  LCD_insert_decimal( 11 );
  }
}

/*-----------------------------------------------------------------
  check_on_or_off()

  See if the appliance should be on or off. Called once a second.

  Returns: 1 (on) or 0 (off)
-----------------------------------------------------------------*/
static uint8_t check_on_or_off()
{
  uint16_t	uiTimeOn, uiTimeOff, uiClock;

  if ( stData.ucActivated != ACTIVATED_YES )
    // Timer is set to deactivated state. Return 0 unless 
	// override is on.
    return ucTempOverrideOn;

  if ( stRtc.rtc_secs == 30 )
  {
    // Arbitrarily at 30 seconds each minute refresh the 
	// preprogram settings.
	set_current_programmed_on_off_times();
  }

  if ( stData.ucPreprogram )
  {
	// Use the on and off times from our table
	uiTimeOn  = ( (uint16_t) ucPrepgmHoursOn * 60 )  + (uint16_t) ucPrepgmMinsOn;
	uiTimeOff = ( (uint16_t) ucPrepgmHoursOff * 60 ) + (uint16_t) ucPrepgmMinsOff;
  }
  else
  {
    // Use the on and off times set by the user
	uiTimeOn  = ( (uint16_t) stData.ucOnTimeHH * 60 )  + (uint16_t) stData.ucOnTimeMM;
	uiTimeOff = ( (uint16_t) stData.ucOffTimeHH * 60 ) + (uint16_t) stData.ucOffTimeMM;
  }

  if ( uiTimeOn == uiTimeOff )
    // On/Off times have not been programmed yet (or improperly programmed)
	// Return 0 unless override is on.
	return ucTempOverrideOn;

  uiClock = ( (uint16_t) stRtc.rtc_hours * 60 ) + (uint16_t) stRtc.rtc_mins;

  // If the on or off time is now and the temporary override is on, cancel it.
  if ( uiClock == uiTimeOn || uiClock == uiTimeOff )
	ucTempOverrideOn = 0;

  // Now see whether the appliance should be on or off
  if ( uiClock >= uiTimeOn || uiClock < uiTimeOff )
	// Return 1 unless temporary override is on
    return ( ucTempOverrideOn ) ? 0 : 1;

  return ucTempOverrideOn;	// return 0 unless temp. override is on
}

/*-----------------------------------------------------------------
  check_DST()

  Called once a second. We need to set the time ahead one hour
  in the spring and one hour back in the fall. Special logic is
  required in the fall to keep us from repeating the hour change.
-----------------------------------------------------------------*/
static void check_DST()
{
  static volatile uint8_t holdoff=0;
  uint8_t    isDST;

  // Do only when seconds and minutes = 0
  if ( stRtc.rtc_secs || stRtc.rtc_mins ) return;

  if ( holdoff )
  {
    // This flag tells us the time was set back one hour recently.
	// Wait until 3AM to prevent us from repeating, then clear the flag.
    if ( stRtc.rtc_hours == 3 )
	  holdoff = 0;

    return;
  }

  isDST = isdaylighttime( stRtc.rtc_year, stRtc.rtc_month, 
                          stRtc.rtc_day,  stRtc.rtc_hours );

  if ( isDST == wasDST ) return;   // no change

  if ( isDST )
  {
    // We have just gone from standard time to DST.
	// We need to put the clock ahead 1 hour.
	++stRtc.rtc_hours;
    rtc_write();
	wasDST = isDST;
  }
  else
  {
    // We have just gone from DST to standard time.
	// We need to put the clock back 1 hour.
	--stRtc.rtc_hours;
    rtc_write();
	wasDST = isDST;
    holdoff = 1;     // to keep us from doing this again in an hour
  }
}

/*-----------------------------------------------------------------
  set_current_programmed_on_off_times()

  Sets the current programmed on and off times from our tables and
  the current clock.

  Output: ucPrepgmHoursOn, ucPrepgmHoursOff, ucPrepgmMinsOn,
		  ucPrepgmMinsOff
-----------------------------------------------------------------*/
static void set_current_programmed_on_off_times()
{
  uint8_t   ucIndex;

  // Table for setting on and off times depending on time of year.
  // To keep it simple there is one entry for each week. All times
  // must be in "standard" time (not DST). These tables are for
  // Toronto, ON (43 deg. 46 min. N, 79 deg. 23 min. W, courtesy
  // http://aa.usno.navy.mil/data.
  static uint8_t ucOntimeHH[52] = 
  {
    16, 16, 16, 17, 17, 17, 17, 17, 17, 17, 18, 18, 18, 
	18, 18, 18, 18, 19, 19, 19, 19, 19, 19, 19, 19, 19,
	19, 19, 19, 19, 19, 19, 19, 19, 18, 18, 18, 18, 17,
	17, 17, 17, 17, 17, 16, 16, 16, 16, 16, 16, 16, 16
  };
  static uint8_t ucOntimeMM[52] = 
  {
    35, 45, 50,  0, 10, 20, 30, 40, 50, 55,  5, 15, 25, 
	30, 40, 50, 55,  5, 15, 20, 30, 35, 40, 45, 50, 50,
	50, 45, 40, 35, 30, 20, 10,  0, 45, 35, 25, 10, 55,
	45, 30, 20, 10, 00, 50, 40, 35, 30, 25, 25, 25, 30
  };
  static uint8_t ucOfftimeHH[52] = 
  {
     8,  8,  8,  8,  7,  7,  7,  7,  7,  7,  6,  6,  6, 
	 6,  6,  5,  5,  5,  5,  5,  5,  4,  4,  4,  4,  4,
	 5,  5,  5,  5,  5,  5,  5,  5,  6,  6,  6,  6,  6,
	 6,  6,  6,  7,  7,  7,  7,  7,  7,  7,  7,  8,  8
  };
  static uint8_t ucOfftimeMM[52] = 
  {
    05, 05,  0,  0, 50, 45, 35, 25, 15, 00, 50, 35, 25, 
	10,  0, 45, 35, 25, 15,  5,  0, 55, 50, 45, 50, 55,
	 0,  5, 10, 15, 25, 35, 40, 50,  0,  5, 15, 20, 30,
	40, 45, 55,  5, 15, 20, 20, 30, 40, 50, 55,  0,  5
  };

  // Use the on and off times from our table
  ucIndex = week_of_year( stRtc.rtc_year, stRtc.rtc_month, stRtc.rtc_day );
  ucPrepgmHoursOn  = ucOntimeHH[ucIndex];
  ucPrepgmHoursOff = ucOfftimeHH[ucIndex];
  ucPrepgmMinsOn   = ucOntimeMM[ucIndex];
  ucPrepgmMinsOff  = ucOfftimeMM[ucIndex];
  if ( isdaylighttime( stRtc.rtc_year, stRtc.rtc_month, stRtc.rtc_day,
						stRtc.rtc_hours ) )
  {
    // Adjust the table values for DST
	++ucPrepgmHoursOn;
	++ucPrepgmHoursOff;
	// It's possible to overflow if hours are 23 but there's no 23
	// in my table. If it's a possibility then test for overflow.
  }
}

/*-----------------------------------------------------------------
  system_init()

  PORTB: PB0 (input) Rotary Encoder "B"
         PB1 (input) Rotary Encoder "A"
         PB2 (input) Rotary Encoder "PB"
         PB3 (input) ISP MOSI
         PB4 (input) ISP MISO
         PB5 (input) ISP SCK
         PB6 (output) not used
         PB7 (output) not used

  PORTC: PC0 (output) RTC data
         PC1 (output) RTC I/O
         PC2 (output) RTC CLK
         PC3 (output) RTC CE
         PC4 (output) diagnostic LED
         PC5 (output) not used
         PC6 (output) ISP !Reset 

  PORTD: PD0 (output) LCD !SCK
         PD1 (output) LCD SI
         PD2 (output) LCD C/!D
         PD3 (output) LCD !Reset
         PD4 (output) LCD !Busy
         PD5 (output) LCD CS
         PD6 (output) LCD Vcc
         PD7 (output) Relay Control
-----------------------------------------------------------------*/
static void system_init()
{
  // The fuses are set to provide an internal clock, running at
  // 8.0 MHz with a prescaler of 8. Change the prescaler from
  // 8 to 1. The procedure requires two steps as per the device
  // specifications.
  CLKPR = 0b10000000;       // set clock prescaler change enable
  CLKPR = 0b00000000;       // set prescaler to 1

  // Set up the I/O ports directions, pull-ups
  DDRB  = 0b11000000;		// pb0..5=input, pb6..7=output
  PORTB = 0b00000111;       // initialize PortB; pull-ups on pb0..2

  DDRC  = 0b11111111;		// pc0..6=output
  PORTC = 0b00000000;       // initialize PortC

  DDRD  = 0b11101111;		// pd0..3=output, pd4=input, pd5..7=output
  PORTD = 0b00101001;       // initialize PortD

  // Initialize Timer1.
  // Timer1 is used to create 1 s intervals for scheduling events
  TCCR1A = 0b00000000;		// normal operation, no OC1A, no OC1B
  TCCR1B = 0b00000100;      // internal clock, prescaler = 256
  TCNT1 = INIT_TIMER1;
  set_bit( TIMSK1, TOIE1 );	// enable timer/counter1 overflow interrupt

  // Initialize Timer2. Timer2 is used to generate interrupts
  // every 1 millisecond for switch debouncing.
  TCCR2A = 0b00000000;		// normal port operation
  TCCR2B = 0b00000011;		// normal operation, prescaler=32
  TCNT2 = INIT_TIMER2;
  set_bit( TIMSK2, TOIE2 );	// enable timer/counter2 overflow interrupt

  sei();					// enable global interrupts
}
